/**
 * \file Bus.cpp
 *
 *  \date Apr 12, 2012
 *  \author Alexandre Tuleu
 */

#include "Bus.h"

#include "LazyWorkflow.h"
#include "ScheduledWorkflow.h"

namespace sbcp {
	Bus::Bus(const sbcpd_bus_settings * settings,
	         sbcpd_serial_interface * si)
		: d_bus(NULL)
		, d_good(false)
		, d_workflow()
		, d_openedDevice(){
		d_bus = sbcpd_create_bus(settings);

		if( d_bus == NULL){
			throw std::runtime_error("unexpected sbcpd error on instantation. Call Aunt Tilda.");
		}

		if(sbcpd_bus_open(d_bus,si) != SBCPD_NO_ERROR){
			throw std::runtime_error(std::string("sbcpd error on bus opening : ") + sbcpd_bus_get_last_error_message(d_bus));
		}

		d_good = true;
	}

	Bus::~Bus(){
		//will release all pending transfer
		d_workflow = BusWorkflow::Ptr();
		sbcpd_bus_close(d_bus);
	}

	bool Bus::Good(){
		return d_good;
	}

	LazyWorkflow & Bus::Lazy(){
		return SwitchToWorkflow<LazyWorkflow>();
	}

	ScheduledWorkflow & Bus::Scheduled(){
		return SwitchToWorkflow<ScheduledWorkflow>();
	}

	Bus::Workflow Bus::CurrentWorkflow() const {
		return d_workflow ? (Workflow)(d_workflow->Type()) : NONE;
	}

	Device::Ptr Bus::OpenGenericDevice(sbcp::Class klass, DevID id){
		if(!Good()){
			throw std::runtime_error("Could not scan device : Bus is not in a good state.");
		}

		//get out if in sheduled mode.
		if(CurrentWorkflow() == SCHEDULED){
			throw std::runtime_error("Could not scan device in scheduled workflow.");
		}

		DevUUID uuid(FromClassID(klass,id));
		DeviceByUUID::iterator fi = d_openedDevice.find(uuid);

		if(fi != d_openedDevice.end()){
			return fi->second;
		}

		LazyWorkflow & w = this->Lazy();

		if(w.Prepare(klass,id,Device::GET_FIRMWARE,0) == NULL){
			return Device::Ptr();
		}

		const Byte * f;
		PayloadSize size;
		Data firm;
		try {
			w.Submit(&f,size);
		} catch (const TransferError & e){
			return Device::Ptr();
		}
		FillDataFromBytes(f,firm);
		w.Release();

		return Device::Ptr(new Device(*this,klass,id,firm));
	}

	bool Bus::SearchForDevice(sbcp::Class klass, DevID id){
		return OpenGenericDevice(klass,id);
	}

} /* namespace sbcp */
