/**
 * \file Bus.h
 *
 *  \date Apr 12, 2012
 *  \author Alexandre Tuleu
 */

#ifndef LIBSBCP_BUS_BUS_H_
#define LIBSBCP_BUS_BUS_H_

#include <tr1/memory>
#include <map>
#include <stdexcept>

#include <sbcpd/sbcpd.h>
#include <sbcpd/sbcp.h>

#include <libsbcp/device/Device.h>

#include "LazyWorkflow.h"



/**
 *
 */
namespace sbcp {
	class ScheduledWorkflow;


	/**
	 * Representation of a physical bus. it :
	 *  - Open and get track of the opened Device 's over the bus.
	 *  - Manages the different Bus::Workflow
	 *
	 *  Basically one would :
	 *  - open the bus, either by construction or using BusConfigSection::OpenBus()
	 *  - open some device using Bus::OpenDevice or Bus::OpenGenericDevice().
	 *  - perform some operation on the Device
	 *
	 *  There is two major BusWorkflow :
	 *  - LazyWorkflow wich is the default and principal mode. Operation are
	 *    performed on by one with a synchronous interface. Any operation could
	 *    be performed.
	 *  - ScheduledWorkflow, which is for fast, low latency operation. Only a
	 *    list of Device is addressed in an asynchronous maneer. Only a
	 *    preselected sets of register are updated.
	 *
	 *
	 * Switch to one fromn the other BusWorkflow is performed using Lazy() or
	 * Scheduled().
	 */
	class Bus {
	public:
		/**
		 * Usable workflow for performing messaging.
		 */
		enum Workflow {
			NONE = 0,LAZY = 1, SCHEDULED = 2};
		//enum WorkflowSwitchMode {THROWING, NON_THROWING};

		Bus(const sbcpd_bus_settings * settings,
		    sbcpd_serial_interface * si);
		virtual ~Bus();

		bool Good();

		/**
		 * @return the current Workflow
		 */
		Workflow CurrentWorkflow() const;

		template <class W>
		W & SwitchToWorkflow();

		/**
		 * Switchs to LazyWorkflow if needed. All remaining I/O are terminated
		 * if needed.
		 * @return a reference to the LazyWorkflow.
		 */
		LazyWorkflow & Lazy();

		/**
		 * Switchs to ScheduledWorkflow if neede. All Remaining I/O operation
		 * are terminated if needed.
		 * @return a reference to the ScheduledWorkflow.
		 */
		ScheduledWorkflow & Scheduled();

		/**
		 * @return a pointer to the unerlying sbcpd_bus.
		 */
		inline sbcpd_bus * SBCPD() {
			return d_bus;
		}

		/**
		 * Finds and open a given Device over the bus.
		 * @tparam DeviceClass the Class of device to open, this is a class
		 *                     defining a const static Class Klass member and
		 *                     inherits Device.
		 * @param id the DevID of the device
		 *
		 * @return a valid DeviceClass::Ptr if found, an invalid one otherwise
		 */
		template <typename DeviceClass>
		std::tr1::shared_ptr<DeviceClass> OpenDevice(DevID id);

		/**
		 * Open a generic device. Use this only if you do not need a specific
		 * interface for the device, and you intend to play only with the
		 * Device::StdRegister
		 * @param klass sbcp::Class of the device
		 * @param id the sbcp::DevID of the deice
		 * @return a Device::Ptr to the device.
		 */
		Device::Ptr OpenGenericDevice(sbcp::Class klass, DevID id);

		/**
		 * Tests of the device is present on the bus.
		 * @param klass Class of the device
		 * @param id DevID of the device
		 * @return true if device is present over the sbcp::Bus.
		 */
		bool SearchForDevice(sbcp::Class klass, DevID id);

	private :
		Bus(const Bus & src); //UNIMPLEMENTED
		void operator=(const Bus & src); //UNIMPLEMENTED
		typedef uint16_t DevUUID;
		typedef std::map<DevUUID,Device::Ptr> DeviceByUUID;


		DevUUID FromClassID(sbcp::Class klass,sbcp::DevID id);

		sbcpd_bus * d_bus;
		bool d_good;

		BusWorkflow::Ptr d_workflow;

		DeviceByUUID d_openedDevice;
	};

	inline Bus::DevUUID Bus::FromClassID(sbcp::Class klass, DevID id){
		return (klass << 8) + id;
	}


	template <class DeviceClass>
	inline std::tr1::shared_ptr<DeviceClass> Bus::OpenDevice(DevID id){
		typedef std::tr1::shared_ptr<DeviceClass> DevPtr;
		if(!Good()){
			throw std::runtime_error("Could not open device : Bus is not in a good state.");
		}

		//get out if in sheduled mode.
		if(CurrentWorkflow() == SCHEDULED){
			throw std::runtime_error("Could not open device in scheduled workflow.");
		}

		//switch to lazy workflow.
		LazyWorkflow & w = this->Lazy();

		DevUUID uuid(FromClassID(DeviceClass::Klass,id));
		DeviceByUUID::iterator fi = d_openedDevice.find(uuid);

		//check if device is not already opened
		if(fi != d_openedDevice.end()){
			return std::tr1::static_pointer_cast<DeviceClass,Device>(fi->second);//DevPtr(static_cast<DeviceClass*>(fi->second.get()));
		}

		if (w.Prepare(DeviceClass::Klass,id,Device::GET_FIRMWARE,0) == NULL){
			return DevPtr();
		}

		const Byte * f;
		PayloadSize size;
		Data firmwareVersion;
		try{
			w.Submit(&f,size);
		} catch (const TransferError & e) {
			return DevPtr();
		}

		FillDataFromBytes(f,firmwareVersion);
		w.Release();

		Device::Ptr device = DevPtr(new DeviceClass(*this,id,firmwareVersion));


		d_openedDevice.insert(std::make_pair(uuid,device));

		return std::tr1::static_pointer_cast<DeviceClass,Device>(device);
	}

	template <class W>
	inline W & Bus::SwitchToWorkflow(){
		if(d_workflow && d_workflow->Type() == W::WORKFLOW){
			return *static_cast<W*>(d_workflow.get());
		}
		d_workflow = W::Creator::Create(*this);
		return *static_cast<W*>(d_workflow.get());
	}
}
#endif // LIBSBCP_BUS_BUS_H_
