/**
 * \file LazyWorkflow.h
 *
 *  \date Apr 12, 2012
 *  \author Alexandre Tuleu
 */

#ifndef LIBSBCP_BUS_LAZYWORKFLOW_H_
#define LIBSBCP_BUS_LAZYWORKFLOW_H_

#include <tr1/memory>

#include <biorob-cpp/base/common.h>

#include <sbcpd/transfer.h>

#include <libsbcp/device/sbcp_common.h>

#include "BusWorkflow.h"

#include <libsbcp/errors/Errors.h>

/**
 *
 */
namespace sbcp {
	class Bus;
	class Device;
	/**
	 * A synchronous, versatile, with high latency BusWorkflow - default one.
	 *
	 * This workflow is used for versatile, on the fly communication over the
	 * bus. It could perform any operation over the bus. Its limitations are :
	 * - Only one device could be addressed by communication loop.
	 * - Device are bound to answer in maximum 1ms, so huge latency coudl happen.
	 * - Interface is synchronous : Submit() will wait until all communication
	 *   is done. Release() only release all internal data of the bus that
	 *   has been allocated for the transfer.
	 * On the other hand there is no limitation on which Device and Instruction
	 * that could be performed.
	 *
	 * To perform a lazy I/O, one should, in that order :
	 * - Prepare() of a given PayloadSize.
	 * - Submit() :  will synchronously perform the transfer over the bus. The
	 *                 current thread will be blocked upon completion.
	 * - Release() : release the response payload data to enable a new
	 *                   transfer.
	 *
	 *  If you fail to follow this order you may get a std::runtime_error.
	 *  Any problem over the SBCP bus may fire SbcdpError. Any problem
	 *  during the transfer may fire TransferError .
	 */
	class LazyWorkflow : public sbcp::BusWorkflow {
		DISABLE_COPY(LazyWorkflow)
		DISABLE_ASSIGNEMENT(LazyWorkflow)
	public:

		//only ::sbcp::Bus could create a BusWorkflow
		class Creator {
		private :
			static BusWorkflow::Ptr Create(Bus & bus);
			friend class ::sbcp::Bus;
		};
		friend class Creator;

		virtual ~LazyWorkflow();

		/**
		 * Prepares a transfer to the target device designed by a Device::Ptr.
		 * It throws std::runtime_error if the LazyWorkflow is not respected, or
		 * SbcpdError on internal bus error.
		 * @param dev teh device to address
		 * @param instruction the InstructionID to perform
		 * @param size the PayloadSize of the transfer
		 * @return the address of an array of size that will hold the payload
		 *         data
		 */
		Byte * Prepare(std::tr1::shared_ptr<Device> dev,
		               InstructionID instruction,
		               PayloadSize size);
		/**
		 *  Prepares a transfer to the target device designed by a sbcp::Class
		 *  and DevID. It throws std::runtime_error if the LazyWorkflow is not
		 *  respected, or SbcpdError on internal bus error.
		 * @param klass sbcp::Class of the device
		 * @param id DevID of teh device
		 * @param instruction InstructionID to perform
		 * @param size PayloadSize of the data to transmit.
		 * @return the address of an array of size that will hold the payload
		 *         data
		 */
		Byte * Prepare(sbcp::Class klass,
		               DevID id,
		               InstructionID instruction,
		               PayloadSize size);

		/**
		 * Performs the transfer of the data over the bus. It throws
		 * std::runtime_error if the LazyWorkflow is not respected, or
		 * SbcpdError on internal bus error. If the transfer is physically
		 * performed but fails for nay reason it throws TransferError.
		 * @param data feedback payload data array address
		 * @param size feedback payload size.
		 */
		void Submit(const Byte ** data,
		                    PayloadSize & size);

		/**
		 * Termninate the transfer. Data that is contained in the array given by
		 * Submit() should not be accessed anymore.
		 */
		void Release();

		const static unsigned int WORKFLOW;

	private :
		//constructor is private
		LazyWorkflow(Bus & bus);

		sbcpd_transfer * d_transfer;
	};

}
#endif // LIBSBCP_LAZYWORKFLOW_H_
