/**
 * \file ScheduledWorkflow.h
 *
 *  \date Apr 12, 2012
 *  \author Alexandre Tuleu
 */

#ifndef LIBSBCP_BUS_SCHEDULEDWORKFLOW_H_
#define LIBSBCP_BUS_SCHEDULEDWORKFLOW_H_

#include <map>

#include <libsbcp/device/Device.h>

#include "BusWorkflow.h"

struct sbcpd_transfer;


namespace sbcp {
	class Bus;
	/**
	 * An asynchronous, low-latency fast BusWorkFlow.
	 *
	 * This BusWorkflow should be used for fast asynchronous I/O over the bus.
	 * Its main drawback is that the list of Device and the list of Register to
	 * be addressed is static and fixed after data has been started to send. But
	 * otherwise it perform well tto make a fast Communication loop over the
	 * bus.
	 *
	 * To perform Scheduled I/O one should :
	 * - AppendScheduledDevice() to select whcih device could be addressed
	 * - StartTransfers() to ignitiate the Transfer
	 * - WaitForTransfersCompletion() to block untill all transfer completion.
	 *
	 * In ScheduledWorkflow, You should not switch to LazyWorkflow
	 * (using Bus::Lazy()), by, for example starting a new Instruction over the
	 * Bus. If you use RegisterAccessor, then RegisterAccessor::Get() and
	 * RegisterAccessor::Set() are safe and will not switch to LazyWorkflow,
	 * but ignore instruction if they could not be performed.
	 *
	 */
	class ScheduledWorkflow : public sbcp::BusWorkflow {
		DISABLE_COPY(ScheduledWorkflow)
		DISABLE_ASSIGNEMENT(ScheduledWorkflow)

	public:
		//only ::sbcp::Bus could create a BusWorkflow
		class Creator {
		private :
			static BusWorkflow::Ptr Create(Bus & bus);
			friend class ::sbcp::Bus;
		};
		friend class Creator;

		virtual ~ScheduledWorkflow();

		/**
		 * Appends a device to the list of scheduled devices.
		 * @param dev Device::Ptr to the Device to schedule.
		 */
		void AppendScheduledDevice(Device::Ptr dev);

		/**
		 * Removes all scheduled Device. All transfers should be terminated if
		 * started using WaitForTransfersCompletion().
		 */
		void ClearAllScheduledDevices();

		/**
		 * Starts the transfer for the scheduled devices.
		 */
		void StartTransfers();

		/**
		 * Block caller until all transfer completion. May throw
		 * MultipleTransferError if any error during tansfer happen.
		 */
		void WaitForTransfersCompletion();


		const static unsigned int WORKFLOW;


	private :
		typedef std::map<Device::Ptr,sbcpd_transfer*> TransferByDevice;

		enum State {
			PREPARATION,
			WAITING_FOR_COMPLETION,
			WAITING_FOR_RESUBMISSION
		};

		ScheduledWorkflow(Bus & bus);

		void ReleaseAllIOTransfers();


		State d_state;

		TransferByDevice d_transfers;
	};
}
#endif // LIBSBCP_SCHEDULEDWORKFLOW_H_
