/**
 * \file device/Device.cpp
 *
 *  \date Mar 31, 2012
 *  \author Alexandre Tuleu
 */

#include "Device.h"

#include <libsbcp/bus/Bus.h>
#include <cstring>
#include <stdexcept>
#include <sstream>
#include <iomanip>

#include <iostream>

namespace sbcp {

	inline void Device::AddRegisterPrimpl(DeviceRegister::Address address,
	                                      uint16_t flags){
		d_registers[address].flags = (DeviceRegister::Flags) (flags & DeviceRegister::PublicFlagsMask);
	}


	Device::Device(Bus & bus,
	               sbcp::Class klass ,
	               DevID id,
	               FirmwareVersion version )
	: d_class(klass)
	, d_id(id)
	, d_version(version)
	, d_registers()
	, d_bus(bus)
	, d_llReaders()
	, d_llWriters(){
		//reserve in the vector for the maximal size of register in the device.
		FillStdRegisters();
	}

	Device::~Device(){
	}

	Device::ListOfRegisters & Device::Registers(){
		return d_registers;
	}

	sbcp::Class Device::Klass() const{
		return d_class;
	}

	DevID Device::ID() const {
		return d_id;
	}

	FirmwareVersion Device::Version() const {
		return d_version;
	}

#define CHECK_ACCES_AND_THROW(a)do{\
	if(a >= d_registers.size()){\
		std::ostringstream os;\
		os << "0x" << std::hex << (int) address\
		   << " is out of bound; device has " << d_registers.size()\
		   << " registers." << std::flush;\
		throw std::overflow_error(os.str());\
	}\
}while(0)


#define ADD_DEVICE_HEADER(dev,os,reg)do{	  \
	os << "Device 0x" << std::hex\
	   << (int) ((Device*) dev )->Klass() << ":0x"\
	   << (int) ((Device*) dev )->ID() << std::dec\
	   << " Register " << (int) reg;\
}while(0)

	void Device::SetRegister(DeviceRegister::Address address, Data data){
//		CHECK_ACCES_AND_THROW(address);

		if(!(d_registers[address].flags & DeviceRegister::WRITABLE)){
			throw std::logic_error("Register is not writable.");
			return;
		}

		if(d_bus.CurrentWorkflow() == Bus::SCHEDULED){
			throw std::runtime_error("The workflow is scheduled, could not have lazy IO.");
		}

		//switch to lazy workflow.
		LazyWorkflow & w = d_bus.Lazy();

		Byte * out = w.Prepare(Klass(),ID(),SET_REGISTERS,3);
		out[0] = address;
		FillBytesFromData(data,&(out[1]));

		const Byte * res(NULL);
		PayloadSize resSize;

		w.Submit(&res,resSize);


		if(resSize != 0 ){
			std::ostringstream os;
			ADD_DEVICE_HEADER(this,os,address);
			os << " received a payload of " << resSize << " for set instruction but none expected.";
			throw std::runtime_error(os.str());
		}

		w.Release();
	}

	Data Device::GetRegister(DeviceRegister::Address address ){
//		CHECK_ACCES_AND_THROW(address);
		DeviceRegister & reg = d_registers[address];
		Data d;
		if(!(reg.flags & DeviceRegister::READABLE)){
			throw std::logic_error("The register is not readable.");
		}
		if(d_bus.CurrentWorkflow() == Bus::SCHEDULED){
			throw std::runtime_error("The workflow is scheduled, could not have lazy IO.");
		}

		//switch to LazyWorkflow
		LazyWorkflow & w = d_bus.Lazy();
		Byte * out = w.Prepare(Klass(),ID(), GET_REGISTERS,1);
		out[0] = address;
		const Byte * res(NULL);
		PayloadSize resSize;
		w.Submit(&res,resSize);

		if( resSize != 2){
			std::ostringstream os;
			ADD_DEVICE_HEADER(this,os,address);
			os << " got " << resSize << " bytes but 2 is expected to get a register";
			throw std::runtime_error(os.str());
		}


		FillDataFromBytes(res,d);
		w.Release();
		return d;
	}

	void Device::FillStdRegisters(){
		d_registers.resize(sbcp::MaxDevRegisters);

		//start by the last, to reduce resize overhead
		AddRegisterPrimpl(SBCP_ID,
		                    DeviceRegister::READABLE
		                  | DeviceRegister::PERSISTENT
		                  | DeviceRegister::WRITABLE);
		AddRegisterPrimpl(FIRMWARE_VERSION,
		                  DeviceRegister::READABLE | DeviceRegister::PERSISTENT);
		AddRegisterPrimpl(SBCP_CLASS,
		                  DeviceRegister::READABLE | DeviceRegister::PERSISTENT);



	}


	void Device::AddRegister(DeviceRegister::Address address, unsigned int flags){
		//DevRegister::Address defintion insure address < MaxDevRegisters
		// this is an uint8_t !
		if(address < DEVICE_SPECIFIC_REGISTERS){
			std::ostringstream os;
			os << std::hex << "Adress 0x" << (int) address
			   << " is reserved for standard registers please "
			   << " choose an address >= 0x" << (int) DEVICE_SPECIFIC_REGISTERS
			   << std::endl;
			throw std::underflow_error(os.str());
		}
		AddRegisterPrimpl(address,flags & DeviceRegister::PublicFlagsMask);
	}


	Bus & Device::GetBus() {
		return d_bus;
	}

	DeviceRegister::Flags
	Device::GetRegisterFlags(DeviceRegister::Address address){
//		CHECK_ACCES_AND_THROW(address);
		return (DeviceRegister::Flags) (d_registers[address].flags & DeviceRegister::PublicFlagsMask);
	}




	void Device::WriteLowLatencyBuffer(Byte* buffer , PayloadSize size){
		PayloadSize expectedSize = 2 * d_llWriters.size();
		if(size != expectedSize){
			std::ostringstream os;
			os << "expecting a transfer output of " << (int) expectedSize
			   << " but got a buffer size of " << (int)size << std::endl;
			throw std::logic_error(os.str());
		}

		for(unsigned int i = 0; i < d_llWriters.size() ; ++i){
			FillBytesFromData(d_llWriters[i]->ToData(),&(buffer[2*i]));
		}
	}

	void Device::ReadLowLatencyBuffer(const Byte* buffer , PayloadSize size){
		PayloadSize expectedSize = 2 * d_llReaders.size();
		if(size != expectedSize){
			std::ostringstream os;
			os << "expecting a transfer input of " << (int) expectedSize
			   << " but got a buffer size of " << (int)size << std::endl;
			throw std::runtime_error(os.str());
		}

		Data d;
		for(unsigned int i = 0; i < d_llReaders.size(); ++i){
			FillDataFromBytes(&(buffer[2 * i]),d);
			d_llReaders[i]->UpdateFromData(d);
		}
	}

	void Device::AppendLowLatencyReadRegister(RegisterReaderBase & reg ){
		d_llReaders.push_back(&reg);
	}

	void Device::AppendLowLatencyWrittenRegister(RegisterWriterBase & reg){
		d_llWriters.push_back(&reg);
		d_registers[reg.Address()].flags = (DeviceRegister::Flags) (d_registers[reg.Address()].flags | DeviceRegister::SCHEDULED_WR);
	}


	void Device::UnlockRegister(DeviceRegister::Address address){
		DeviceRegister & reg = d_registers[address];
		if(! (reg.flags & (DeviceRegister::WRITABLE | DeviceRegister::PERSISTENT))){
			std::ostringstream os;
			os << "0x" << std::hex << reg.flags << " against " << "0x" << (int) (DeviceRegister::WRITABLE | DeviceRegister::PERSISTENT);
			throw std::logic_error("You can only unlock WRITABLE && PERSISTENT registers (" + os.str() + ")");
		}

		if(d_bus.CurrentWorkflow() != Bus::LAZY){
			throw std::runtime_error("Unlocking a register could only happen in Lazy workflow");
		}

		LazyWorkflow & w = d_bus.Lazy();

		Byte * out = w.Prepare(this->Klass(),this->ID(),UNLOCK_REGISTER,1);

		out[0] = address;

		const Byte * res(NULL);
		PayloadSize resSize;

		w.Submit(&res,resSize);

		w.Release();
		if( resSize != 0){
			std::ostringstream os;
			ADD_DEVICE_HEADER(this,os,address);
			os << " got " << resSize << " bytes but none is expected.";
			throw std::runtime_error(os.str());
		}
		return;

	}




}/* namespace sbcp */
