/**
 * \file device/Device.h
 *
 *  \date Mar 31, 2012
 *  \author Alexandre Tuleu
 */

#ifndef LIBSBCP_DEVICE_DEVICE_H_
#define LIBSBCP_DEVICE_DEVICE_H_


#include <vector>
#include <tr1/memory>

#include <libsbcp/device/sbcp_common.h>
#include "RegisterAccessorBase.h"


namespace sbcp{


	class Bus;
	/**
	 * Represents a generic Device over the Bus.
	 *
	 * \note Device are not created directly but are rather instantiated using
	 *       sbcp::Bus::OpenDevice() template function.
	 *
	 * Please notice, that most of the time you want to get a specific
	 * implementation of the device, like sbcp::amarsi::MotorDriver or
	 * sbcp::amarsi::Power.
	 *
	 * A Device as several caracteristics :
	 * - It has an UID over the bus made from a sbcp::Class 8-bit Byte and a
	 *   sbcp::DevID 8-bit Byte.
	 * - It has a table of DeviceRegister that hold some cobtrol data.
	 * - It has a set of standard instruction that could be performed.
	 *
	 * Each device has the same sets of StdInstruction and StdRegister. This class
	 * Implements basic helper method to perfrom these instructions, like
	 * SetRegister(), GetRegister(), UnlockRegister().
	 */
	class Device {
	public:
		/**
		 * shared_ptr typedef
		 */
		typedef std::tr1::shared_ptr<Device> Ptr;

		/**
		 * Binary 8-bit code for InstructionID
		 */
		enum StdInstruction {
			GET_FIRMWARE   = 0,             //!< GET_FIRMWARE  returns the 16-bit FirmwareVersion code of the device.
			GET_REGISTERS = 1,              //!< GET_REGISTERS returns the value of a list of registers.
			SET_REGISTERS = 2,              //!< SET_REGISTERS sets the value of a list of registers
			UNLOCK_REGISTER = 3,            //!< UNLOCK_REGISTER unlocks a list of register
			DEVICE_SPECIFIC_INSTRUCTIONS = 4//!< DEVICE_SPECIFIC_INSTRUCTIONS start of address for device specific instruction
		};

		/**
		 * Binary 8-bit code for DeviceRegister. A Device Register is a 16-bit value.
		 */
		enum StdRegister {
			FIRMWARE_VERSION = 0,        //!< FIRMWARE_VERSION FirmwareVersion code
			SBCP_CLASS       = 1,        //!< SBCP_CLASS       sbcp::Class 8-bit value
			SBCP_ID          = 2,        //!< SBCP_ID          sbcp::DevID 8-bit value
			DEVICE_SPECIFIC_REGISTERS = 3//!< DEVICE_SPECIFIC_REGISTERS start for device-specific DeviceRegister
		};


		Device(Bus & bus,
		       Class klass ,
		       DevID id,
		       FirmwareVersion version);
		virtual ~Device();

		/**
		 * @return the sbcp::Class of the Device.
		 */
		sbcp::Class Klass() const;
		/**
		 * @return the sbcp::DevID of the device.
		 */
		DevID ID() const;
		/**
		 * @return the FirmwareVersion of the device
		 */
		FirmwareVersion Version() const;

		//way to handle instruction

		/**
		 * Sets the valeu of one single register.
		 * Performs a LazyWorkflow transfer.
		 * @param address address of the DeviceRegister.
		 * @param data new value of the DeviceRegister
		 */
		void SetRegister(DeviceRegister::Address address, Data data);

		/**
		 * Gets the value of a DeviceRegister.
		 * Performs a LazyWorkflow transfer.
		 * @param address address of the DeviceRegister.
		 * @return the current value of the DeviceRegister
		 */
		Data GetRegister(DeviceRegister::Address address);

		/**
		 * Unlocks the DeviceRegister. Performs a LazyWorkflow Transfer.
		 * @param address
		 */
		void UnlockRegister(DeviceRegister::Address address);


		sbcp::Bus & GetBus();

		DeviceRegister::Flags GetRegisterFlags(DeviceRegister::Address address);


		void WriteLowLatencyBuffer(Byte * buffer, PayloadSize size);

		void ReadLowLatencyBuffer(const Byte * buffer, PayloadSize size);

		class Accessor{
			static void AddRegister(Device & d,
			                        DeviceRegister::Address address,
			                        unsigned int flags);
			template <typename T,typename InputType> friend class RegisterAccessorBase;
		};
	protected :
		typedef std::vector<DeviceRegister> ListOfRegisters;

		ListOfRegisters & Registers();



		void AppendLowLatencyReadRegister(RegisterReaderBase & reg);

		void AppendLowLatencyWrittenRegister(RegisterWriterBase & reg);



	private :

		friend class Accessor;

		void AddRegister(DeviceRegister::Address address, unsigned int flags);


		typedef std::vector<RegisterWriterBase*> ListOfWriters;
		typedef std::vector<RegisterReaderBase*> ListOfReaders;


		void FillStdRegisters();

		inline void AddRegisterPrimpl(DeviceRegister::Address address,
		                              uint16_t flags);



		sbcp::Class     d_class;
		DevID           d_id;
		FirmwareVersion d_version;

		std::vector<DeviceRegister> d_registers;

		sbcp::Bus & d_bus;

		ListOfReaders d_llReaders;
		ListOfWriters d_llWriters;

	};


}


inline
void sbcp::Device::Accessor::AddRegister(Device & d,
                                         DeviceRegister::Address address,
                                         unsigned int flags) {
	d.AddRegister(address,flags);
}



#endif // LIBSBCP_DEVICE_DEVICE_H_
