/**
 * \file device/RegisterAccessor.h
 *
 *  \date Apr 30, 2012
 *  \author Alexandre Tuleu
 */

#ifndef LIBSBCP_DEVICE_REGISTERACCESSOR_H_
#define LIBSBCP_DEVICE_REGISTERACCESSOR_H_

#include <sstream>
#include <iomanip>

#include <libsbcp/bus/Bus.h>

#include "RegisterAccessorBase.h"

/**
 *
 */
namespace sbcp{

/**
 * Base class for RegisterReader, RegisterWriter and RegisterAccessor.
 *
 * \tparam T : the type of the data for the user. The data of the register will
 *             automatically be converted to that type.
 * \tparam InpuType the input type of the data. Should be 16-bit bit size.
 *         Most ofen it is just a 16-bit int, signed or unsigned.

 */
template <typename T, typename InputType = sbcp::Data>
class RegisterAccessorBase {
public:
	RegisterAccessorBase(Device  & dev, DeviceRegister::Address address,
	                     unsigned int flags);
	virtual ~RegisterAccessorBase();

	/**
	 * @return last writen or read value.
	 */
	T LastValue() const;
	const static unsigned int AdditionalFlagsMask = DeviceRegister::PERSISTENT;
protected :

	Device &                d_dev;
	const DeviceRegister::Address d_address;
	T d_value;
};


/**
 * An object that could read a value of a Device register over the bus.
 *
 * They are usually created by Device class implementation constructor.
 * A end-user should not create them.
 *
 * They provide read access to the data through Get().
 *
 *
 */
template <typename T, typename InputType = sbcp::Data>
class RegisterReader : public virtual RegisterAccessorBase<T>,
                       public RegisterReaderBase {
public :
	typedef std::tr1::shared_ptr<RegisterReader<T,InputType> > Ptr;
	RegisterReader(Device & dev, DeviceRegister::Address address,
	               unsigned int additionalFlags = 0);
	virtual ~RegisterReader();

	/**
	 * @return A new Read of the value. Will perform an hardware tansfer if in
	 *  LazyWorkflow, so the thread may be blocked for a long time. In
	 *  ScheduledWorkflow returns the most updated value.
	 */
	T Get();

protected :
	virtual T FromInternal(InputType value);

private :
	typedef RegisterAccessorBase<T> Base;
	void UpdateFromData(Data d);
};

/**
 * An object that could write to a Device Register.
 *
 * They are usually created by Device class implementation constructor.
 * A end-user should not create them.
 *
 * They provide write access to the data through Set().
 */
template <typename T, typename InputType = sbcp::Data>
class RegisterWriter : public virtual RegisterAccessorBase<T>,
                       public RegisterWriterBase {
public :
	typedef std::tr1::shared_ptr<RegisterWriter<T,InputType> > Ptr;
	RegisterWriter(Device & dev, DeviceRegister::Address address,
	               unsigned int additionalFlags = 0);
	virtual ~RegisterWriter();

	/**
	 * Write the value to the Register. In LazyWorkflow, it will perform a
	 * hardware transfer, so the calling thread may be blocked. Otherwise the
	 * internal data is updated but no transfer performs.
	 * @param value The new value to write.
	 */

	void Set(T value);

	/**
	 * Unlocks the register for writing to a persistent register.
	 */
	void Unlock();

	DeviceRegister::Address Address() const;

protected :
	virtual InputType FromExternal(T value);

private :
	typedef RegisterAccessorBase<T> Base;
	InputType ToData();
};


/**
 * An object that could read and write to a Device register.
 *
 * They are usually created by Device class implementation constructor.
 * A end-user should not create them.
 *
 * I/O could be performed using Set() and Get().
 */
template <typename T, typename InputType = sbcp::Data>
class RegisterAccessor : public RegisterWriter<T>, public RegisterReader<T>{
public :
	typedef std::tr1::shared_ptr<RegisterAccessor<T,InputType> > Ptr;
	RegisterAccessor(Device & dev, DeviceRegister::Address address,
	                 unsigned int additionalFlags = 0);
	virtual ~RegisterAccessor();
};


/*******************************************************************************
 * INLINE IMPLEMENTATION
 ******************************************************************************/


//RegisterAccessorBase


template <typename T, typename InpuType>
inline RegisterAccessorBase<T,InpuType>::RegisterAccessorBase(Device & dev,
                                                              DeviceRegister::Address address,
                                                              unsigned int flags)
	: d_dev(dev)
	, d_address(address)
	, d_value(){
	Device::Accessor::AddRegister(dev,address,flags);
}



template <typename T, typename InpuType>
inline RegisterAccessorBase<T,InpuType>::~RegisterAccessorBase(){}



template <typename T, typename InpuType>
T RegisterAccessorBase<T,InpuType>::LastValue() const{
	return d_value;
}

//RegisterReader


template <typename T, typename InputType>
inline RegisterReader<T,InputType>::RegisterReader(Device  & dev,
                                                   DeviceRegister::Address address,
                                                  unsigned int additionalFlags)
	: RegisterAccessorBase<T>(dev,
	                          address,
	                          DeviceRegister::READABLE |
	                          (additionalFlags & RegisterAccessorBase<T>::AdditionalFlagsMask)){
}



template <typename T, typename InpuType>
inline RegisterReader<T,InpuType>::~RegisterReader(){}



template <typename T, typename InputType>
inline T RegisterReader<T,InputType>::Get(){
	if(Base::d_dev.GetBus().CurrentWorkflow() == Bus::SCHEDULED){
		return Base::d_value;
	}


	Base::d_value = FromInternal( Base::d_dev.GetRegister(Base::d_address));
	return Base::d_value;
}



template <typename T, typename InputType>
inline T RegisterReader<T,InputType>::FromInternal(InputType value){
	return (T) value;
}



template <typename T, typename InputType>
inline void RegisterReader<T,InputType>::UpdateFromData(Data d){
	Base::d_value = FromInternal((InputType)d);
}


//RegisterWriter


template <typename T, typename InputType>
inline RegisterWriter<T,InputType>::RegisterWriter(Device  & dev,
                                                   DeviceRegister::Address address,
                                                   unsigned int additionalFlags)
	: RegisterAccessorBase<T>(dev,
	                          address,
	                          DeviceRegister::WRITABLE |
	                          (additionalFlags & RegisterAccessorBase<T>::AdditionalFlagsMask)){
}



template <typename T, typename InputType>
inline RegisterWriter<T,InputType>::~RegisterWriter(){}



template <typename T, typename InputType>
inline void RegisterWriter<T,InputType>::Set(T value){
	Base::d_value = value;

	if(Base::d_dev.GetBus().CurrentWorkflow() == Bus::SCHEDULED ){
		return;
	}

	Base::d_dev.SetRegister(Base::d_address,FromExternal(Base::d_value));
}

template <typename T, typename InputType>
inline void RegisterWriter<T,InputType>::Unlock() {
	Base::d_dev.UnlockRegister(Base::d_address);
}

template <typename T, typename InputType>
inline DeviceRegister::Address RegisterWriter<T,InputType>::Address() const{
	return Base::d_address;
}

template <typename T, typename InputType>
inline InputType RegisterWriter<T,InputType>::ToData(){
	return FromExternal(Base::d_value);
}



template <typename T, typename InputType>
inline InputType RegisterWriter<T,InputType>::FromExternal(T value){
	return (Data)value;
}


//RegisterAccessor


template <typename T, typename InputType>
inline RegisterAccessor<T,InputType>::RegisterAccessor(Device  & dev,
                                                       DeviceRegister::Address address,
                                                       unsigned int additionalFlags)
	: RegisterAccessorBase<T>(dev,
	                          address,
	                          DeviceRegister::READABLE |
	                          DeviceRegister::WRITABLE |
	                          (additionalFlags & RegisterAccessorBase<T>::AdditionalFlagsMask))
	, RegisterWriter<T>(dev,address)
	, RegisterReader<T>(dev,address)
{
}

template <typename T, typename InputType>
inline RegisterAccessor<T,InputType>::~RegisterAccessor(){}




}
#endif // LIBSBCP_DEVICE_REGISTERACCESSOR_H_
