/**
 * \file device/sbcp_common.h
 *
 *  \date Oct 6, 2011
 *  \author Alexandre Tuleu
 */

#ifndef LIBSBCP_DEVICE_SBCP_COMMON_H_
#define LIBSBCP_DEVICE_SBCP_COMMON_H_

#include <stdint.h>

/**
 * \brief Principal namespace for libsbcp.
 *
 * All devices and registered classe lies under this namespace.
 */

namespace sbcp{
	const static uint8_t MaxDevRegisters = 0xff;

	/**
	 * \brief Type of a sbcp::Device .
	 *
	 * Each Device over the network has a class, that defines what its purpose,
	 * and its protocol interface for early implementation. In the future,
	 * communication interfaces will be identified with the tupple (id,
	 * firmware version)
	 */
	typedef uint8_t Class;
	/**
	 * \brief Id of the sbcp::Device .
	 *
	 * Each sbcp::Device over the Bus has an unique ID byte over the bus. This
	 * allow us to target a specific device.
	 *
	 */
	typedef uint8_t DevID;
	/**
	 * \brief Transfer instruction type
	 *
	 * Each transfer over the sbcp::Bus (to a specific sbcp::Device) contains an
	 * Instruction byte that define the action that will ignitiate this transfer
	 * on the sbcp::Device
	 */
	typedef uint8_t InstructionID;
	/**
	 * \brief Error code type
	 *
	 * Each transfer over the sbcp::Bus will return an ErrorCode. 0 means no
	 * error.
	 */
	typedef uint8_t ErrorCode;
	/**
	 * \brief Size of the payload for the transfer.
	 *
	 * Each transfer may contains a payload of data. this is the size of this
	 * payload in Bytes.
	 */
	typedef uint8_t PayloadSize;

	/**
	 * \brief Byte type.
	 * Self documented portable type.
	 */
	typedef uint8_t Byte;
	/**
	 * \brief Payload type.
	 *
	 * SBCP right now is at the moment targetting mostly 16 bits MCU.
	 * Therefor the size of a DeviceRegister data is 16 bits.
	 *
	 */
	typedef uint16_t Data;

	/**
	 * \brief Version of the sbcp::Device Firmware
	 *
	 * Each Device should provide their firmware version.
	 */
	typedef Data FirmwareVersion;

	/**
	 * Represents a sbcp::Device 16 bits register.
	 *
	 * They are here to have a generic definition of what DeviceRegister a device
	 * has, and what properties do they have.
	 */
	struct DeviceRegister{
	public:
		/**
		 * Flags of the register.
		 */
		enum Flags{
			INEXISTANT     = 0,     //!< INEXISTANT internal initialization state.
			READABLE       = 1 << 0,//!< READABLE this register is readable outside of
			                        //!<          of the sbcp::Device
			WRITABLE       = 1 << 1,//!< WRITABLE this register is writable outside of
			                        //!<          of the sbcp::Device .
			PROTECTED      = 1 << 2,//!< PROTECTED unimplemented
			PERSISTENT     = 1 << 3,//!< PERSISTENT unimplemented or even unuseful at our level
			SCHEDULED_WR   = 1 << 4 //!< SCHEDULED_WR scheduled for write, internal only value.
		};

		const static Flags PublicFlagsMask = (Flags) 0x1F;
		typedef uint8_t Address;

		Flags   flags;
		//Address address;
		DeviceRegister() : flags(INEXISTANT){};
	};

	/**
	 * \private
	 * Not part of the API.
	 */
	inline void FillDataFromBytes(const Byte * b , Data & d){
		d = b[0] << 8;
		d+= b[1];
	}

	/**
	 * \private
	 * Not part of the API
	 */
	inline void FillBytesFromData(const Data & d, Byte *b){
		b[0] = (d & 0xff00) >> 8;
		b[1] =  d & 0x00ff;
	}

}


#endif // LIBSBCP_DEVICE_SBCP_COMMON_H_
