/**
 * \file Power.cpp
 *
 *  \date Oct 30, 2012
 *  \author Alexandre Tuleu
 */

#include "Power.h"

namespace sbcp{
namespace amarsi {

Power::Power(Bus & bus,DevID id, FirmwareVersion version)
	: Device(bus,Klass,id,version)
	, d_voltageMax(*this, VOLTAGE_MAX, DeviceRegister::PERSISTENT)
	, d_voltageMin(*this, VOLTAGE_MIN, DeviceRegister::PERSISTENT)
	, d_voltageCritical(*this, VOLTAGE_CRITICAL, DeviceRegister::PERSISTENT)
	, d_currentMax(*this, CURRENT_MAX, DeviceRegister::PERSISTENT)
	, d_speakerVolume(*this,SPEAKER_VOLUME)
	, d_speakerFrequency(*this,SPEAKER_FREQUENCY)
	, d_inputVoltage(*this,POWER_INPUT_VOLTAGE)
	, d_inputCurrent(*this,POWER_INPUT_VOLTAGE)
	, d_status(*this,POWER_STATUS)
	, d_monitorMode(*this, POWER_MONITOR_MODE)
{
}

Power::~Power(){
}


void Power::AnnounceHostIsUp(){
	LazyWorkflow & w = GetBus().Lazy();


	w.Prepare(Klass,ID(), PWB_HOST_IS_UP,0);
	const Byte * data;
	sbcp::PayloadSize s;
	w.Submit(&data,s);
	w.Release();
	if(s != 0){
		throw std::runtime_error("Unexpected feedback size on AnnouceHostIsUp instruction");
	}
}

bool Power::ShouldShutDown(){
	return (d_status.Get() & (ABOVE_CRITICAL | UNDER_MIN | UNDER_CRITICAL | HOST_SHUTDOWN_PENDING));
}

bool Power::LogicIsUp(){
	return (d_monitorMode.Get() & (ENABLE_LOGIC));
}

bool Power::ServoIsUp(){
	return (d_monitorMode.Get() & (ENABLE_SERVO));
}

bool Power::MotorIsUp(){
	return (d_monitorMode.Get() & (ENABLE_MOTOR));
}

bool Power::IsMonitoringCurrent(){
	return (d_monitorMode.Get() & (MONITOR_CURRENT));
}

void Power::SetActuatorUp(bool enableServo , bool enableMotor){
	MonitorMode m = d_monitorMode.Get();
	if(enableMotor){
		m = (MonitorMode)(ENABLE_MOTOR | m);
	} else {
		m = (MonitorMode)( m & (~ENABLE_MOTOR));
	}

	if(enableServo){
		m = (MonitorMode)(ENABLE_SERVO | m);
	} else {
		m = (MonitorMode)(m & (~ENABLE_SERVO));
	}
	d_monitorMode.Set(m);
}

void Power::SetCurrentMonitoring(bool enable){
	MonitorMode m = d_monitorMode.Get();
	if(enable){
		m = (MonitorMode)(m | MONITOR_CURRENT);
	} else {
		m = (MonitorMode)(m & (~MONITOR_CURRENT));
	}
	d_monitorMode.Set(m);
}

RegisterAccessor<Data>& Power::VoltageMax(){
	return d_voltageMax;
}

RegisterAccessor<Data>& Power::VoltageMin(){
	return d_voltageMin;
}

RegisterAccessor<Data>& Power::VoltageCritical(){
	return d_voltageCritical;
}

RegisterAccessor<Data>& Power::CurrentMax(){
	return d_currentMax;
}

RegisterReader<Power::StatusFlags>& Power::Status(){
	return d_status;
}

RegisterAccessor<Power::MonitorMode>& Power::Monitor(){
	return d_monitorMode;
}

RegisterReader<Data>& Power::InputVoltage(){
	return d_inputVoltage;
}

RegisterReader<Data>& Power::InputCurrent(){
	return d_inputCurrent;
}

RegisterAccessor<Data>& Power::SpeakerVolume(){
	return d_speakerVolume;
}

RegisterAccessor<Data>& Power::SpeakerFrequency(){
	return d_speakerFrequency;
}


} //namespace amarsi
} //namespace sbcp
