/**
 * \file Power.h
 *
 *  \date Oct 30, 2012
 *  \author Alexandre Tuleu
 */

#ifndef LIBSBCP_DEVICE_AMARSI_POWER_H_
#define LIBSBCP_DEVICE_AMARSI_POWER_H_

#include <libsbcp/device/Device.h>
#include <libsbcp/device/RegisterAccessor.h>



namespace sbcp {

namespace amarsi {
/**
 * Driver class for AMARSi Power board SBCP device.
 */
class Power : public Device {
public:
	/**
	 * Device register of the board.
	 */
	enum Registers {
		VOLTAGE_MAX          = Device::DEVICE_SPECIFIC_REGISTERS + 0,//!< VOLTAGE_MAX
		VOLTAGE_MIN          = Device::DEVICE_SPECIFIC_REGISTERS + 1,//!< VOLTAGE_MIN
		VOLTAGE_CRITICAL     = Device::DEVICE_SPECIFIC_REGISTERS + 2,//!< VOLTAGE_CRITICAL
		CURRENT_MAX          = Device::DEVICE_SPECIFIC_REGISTERS + 3,//!< CURRENT_MAX
		POWER_STATUS         = Device::DEVICE_SPECIFIC_REGISTERS + 4,//!< POWER_STATUS
		POWER_MONITOR_MODE   = Device::DEVICE_SPECIFIC_REGISTERS + 5,//!< POWER_MONITOR_MODE
		POWER_INPUT_VOLTAGE  = Device::DEVICE_SPECIFIC_REGISTERS + 6,//!< POWER_INPUT_VOLTAGE
		POWER_INPUT_CURRENT  = Device::DEVICE_SPECIFIC_REGISTERS + 7,//!< POWER_INPUT_CURRENT
		SPEAKER_FREQUENCY    = Device::DEVICE_SPECIFIC_REGISTERS + 8,//!< SPEAKER_FREQUENCY
		SPEAKER_VOLUME       = Device::DEVICE_SPECIFIC_REGISTERS + 9 //!< SPEAKER_VOLUME
	};

	/**
	 * Specific instructions.
	 */
	enum Instructions {
		PWB_HOST_IS_UP = Device::DEVICE_SPECIFIC_INSTRUCTIONS + 0//!< PWB_HOST_IS_UP
	};

	/**
	 * Input voltage status
	 */
	enum StatusFlags {
		NULL_STATUS           = 0,       //!< NULL_STATUS
		ABOVE_CRITICAL        = (1 << 0),//!< ABOVE_CRITICAL
		ABOVE_MIN             = (1 << 1),//!< ABOVE_MIN
		UNDER_MIN             = (1 << 2),//!< UNDER_MIN
		UNDER_CRITICAL        = (1 << 3),//!< UNDER_CRITICAL
		OVER_CURRENT          = (1 << 4),//!< OVER_CURRENT
		HOST_SHUTDOWN_PENDING = (1 << 5) //!< HOST_SHUTDOWN_PENDING
	};

	/**
	 * Enabling / Disabling of power line
	 */
	enum MonitorMode {
		DISABLE_ALL     = 0x00,    //!< DISABLE_ALL
		ENABLE_LOGIC    = (1 << 0),//!< ENABLE_LOGIC
		ENABLE_MOTOR    = (1 << 1),//!< ENABLE_MOTOR
		ENABLE_SERVO    = (1 << 2),//!< ENABLE_SERVO
		MONITOR_CURRENT = (1 << 3) //!< MONITOR_CURRENT
	};

	const static sbcp::Class Klass = 0xeb;
	Power(Bus & bus,DevID id, FirmwareVersion version);
	virtual ~Power();

	/**
	 * Announce the board that the host is up. Switch to LazyWorkflow.
	 */
	void AnnounceHostIsUp();

	/**
	 * @return true if the logic line is about to be shut down
	 */
	bool ShouldShutDown();

	/**
	 * @return true if the logic line is up (should always be the case :  your computer is plugged on it !!)
	 */
	bool LogicIsUp();
	/**
	 * @return true if the servo line is up
	 */
	bool ServoIsUp();
	/**
	 * @return true if the brushless motor line is up
	 */
	bool MotorIsUp();
	/**
	 * @return true if the board try to monitor current
	 */
	bool IsMonitoringCurrent();

	/**
	 * Enable the power line for the servo
	 * @param enableServo if true enable servo, otherwise disable it
	 * @param enableMotor if true enable brushless motor, otehrwise disable it
	 */
	void SetActuatorUp(bool enableServo = true, bool enableMotor = true);
	/**
	 * Sets the monitoring of teh current.
	 * @param enable if true, enable it otherwise disable
	 */
	void SetCurrentMonitoring(bool enable = false);

	/**
	 * Maximal admissible input voltage. Persistent
	 * @return RegisterAccessor
	 */
	RegisterAccessor<Data> & VoltageMax();
	/**
	 * Minimal admissible input voltage. Persitent. If above, logic will be asked gently
	 * to shutdown.
	 * @return RegisterAccessor
	 */
	RegisterAccessor<Data> & VoltageMin();
	/**
	 * Critical Voltage threshold. If above, all linbe will be disabled ASAP.
	 * @return RegisterAccessor
	 */
	RegisterAccessor<Data> & VoltageCritical();
	/**
	 * Maximal admissible current. If IsMonitoringCurrent() and above, all Line shutdown immediatly.
	 * @return
	 */
	RegisterAccessor<Data> & CurrentMax();

	/**
	 * Current board status
	 * @return RegisterReader
	 */
	RegisterReader<StatusFlags> & Status();

	RegisterAccessor<MonitorMode> & Monitor();

	RegisterReader<Data> & InputVoltage();
	RegisterReader<Data> & InputCurrent();

	/**
	 * Speaker Volume.
	 * @return RegisterAccessor
	 */
	RegisterAccessor<Data> & SpeakerVolume();
	/**
	 * Speaker Frequency in Hz.
	 * @return
	 */
	RegisterAccessor<Data> & SpeakerFrequency();



private :
	RegisterAccessor<Data> d_voltageMax,d_voltageMin,d_voltageCritical,
	                            d_currentMax,d_speakerVolume,d_speakerFrequency;
	RegisterReader<Data>          d_inputVoltage,d_inputCurrent;
	RegisterReader<StatusFlags>   d_status;
	RegisterAccessor<MonitorMode> d_monitorMode;

};

}
}

#endif // LIBSBCP_DEVICE_AMARSI_POWER_H_
