/**
 * \file TransferError.cpp
 *
 *  \date May 4, 2012
 *  \author Alexandre Tuleu
 */

#include "TransferError.h"

#include <sbcpd/sbcp.h>

#include <sstream>

namespace sbcp {

	TransferError::TransferError(sbcpd_transfer * t)
		: std::runtime_error("")
		, d_emettorClass(0)
		, d_destClass(t->out_buffer[SBCP_CLASS_POS])
		, d_emmettorID(0)
		, d_destID(t->out_buffer[SBCP_ID_POS])
		, d_feedbackByteValue(0)
		, d_instruction(t->out_buffer[SBCP_INSTRUCTION_POS])
		, d_type(Communication)
		, d_payloadData()
		, d_message(){
		std::ostringstream os;
		d_destClass = t->out_buffer[SBCP_CLASS_POS];
		if(t->in_buffer == NULL){
			os << "Could not read sufficient data from the bus when addressing"
			   << "device 0x" << std::hex << (int) d_destClass << ":0x"
			   << (int) d_destID ;
			d_message = os.str();
			return;
		}

		d_emettorClass = t->in_buffer[SBCP_CLASS_POS];
		d_emmettorID = t->in_buffer[SBCP_ID_POS];
		d_feedbackByteValue = t->in_buffer[SBCP_INSTRUCTION_POS];
		d_type = d_feedbackByteValue & 0x80 ? Communication : Application;
		d_payloadData.reserve(t->in_buffer[SBCP_PAYLOAD_SIZE_POS]);

		for(unsigned int i = 0; i< t->in_buffer[SBCP_PAYLOAD_SIZE_POS]; ++i){
			d_payloadData.push_back(t->in_buffer[SBCP_PAYLOAD_START_POS + i]);
		}


		os << "Got "
		   << ( Type() == TransferError::Application ? "application"
		                                             : "communication")
		   << " error during transfer (status : 0x" << std::hex
		   << (int) t->status <<" com error : 0x" << (int) t->com_error
		   << ") to Device 0x"
		   << (int) DestinationClass() << ":0x"
		   << (int) DestinationID() << " instruction 0x"
		   << (int) Instruction()
		   << " : Device 0x" << (int) EmettorClass()
		   << ":0x" << (int) EmettorID() << " issued error : "
		   << (int) FeedbackByte() << " , payload : [ ";
		for(std::vector<Byte>::const_iterator pb = PayloadData().begin();
		    pb != PayloadData().end();
		    ++pb){
			if(pb != PayloadData().begin()){
				os << " , " ;
			}
			os << "0x" << (int) *pb;
		}
		os << " ]";

		d_message = os.str();
	}

	TransferError::~TransferError() throw (){
	}

	const char * TransferError::what() const throw() {
		return d_message.c_str();
	}


	sbcp::Class TransferError::DestinationClass() const{
		return d_destClass;
	}

	sbcp::DevID TransferError::DestinationID() const{
		return d_destID;
	}

	InstructionID TransferError::Instruction() const{
		return d_instruction;
	}

	TransferError::ErrorType TransferError::Type() const{
		return d_type;
	}

	sbcp::Class TransferError::EmettorClass() const{
		return d_emettorClass;
	}

	sbcp::DevID TransferError::EmettorID() const{
		return d_emmettorID;
	}

	Byte TransferError::FeedbackByte() const{
		return d_feedbackByteValue;
	}

	const std::vector<Byte>& TransferError::PayloadData() const{
		return d_payloadData;
	}

	MultipleTransferError::MultipleTransferError()
		: std::runtime_error("")
		, d_errors()
		, d_message("No errors"){
	}

	MultipleTransferError::~MultipleTransferError() throw(){}

	const char * MultipleTransferError::what() const throw(){
		return d_message.c_str();
	}

	const MultipleTransferError::ListOfError & MultipleTransferError::Errors() const{
		return d_errors;
	}

	void MultipleTransferError::AppendError(const TransferError & e){
		if(d_errors.empty()){
			d_message.clear();
		}
		d_errors.push_back(e);
		d_message.append(e.what() + std::string("\n"));
	}
} /* namespace sbcp */
